/**************************************************************************
 *
 * FILE  make_tapmaster.c
 * Copyright (c) 2012, 2013 Daniel Kahlin
 * Written by Daniel Kahlin <daniel@kahlin.net>
 *
 * DESCRIPTION
 *   Create native tape mastering binary from .tap.
 *
 ******/
#include <stdarg.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>

#include "tapmaster.h"

#define PROGRAM "make_tapmaster"
#define VERSION "0.4"

int verbose_g;
int debug_g;
int force_g;

void panic(const char *str, ...)
{
    va_list args;

    fprintf(stderr, "%s: ", PROGRAM);
    va_start(args, str);
    vfprintf(stderr, str, args);
    va_end(args);
    fputc('\n', stderr);
    exit(1);
}

time_t file_modified(const char *name);
void tapfix(const char *srcname, const char *destname);
void exomize(const char *srcname, const char *destname);

char *outname_g;
char *title_g[16];
int titlenum_g;

int main(int argc, char *argv[])
{
    int c;
    char *srcname;
    FILE *fp;

    /* defaults */
    verbose_g = 0;
    debug_g = 0;
    outname_g = "a.out";
    force_g = 0;
    titlenum_g = 0;

    /*
     * scan for valid options
     */
    while (EOF != (c = getopt(argc, argv, "t:fo:dvVh"))) {
        switch (c) {
	
	/* a missing parameter */
	case ':':
	/* an illegal option */
	case '?':
	    exit(1);

	/* set verbose */
	case 'v':
	    verbose_g = 1;
	    break;

	/* set debug */
	case 'd':
	    debug_g = 1;
	    break;

	/* print version */
	case 'V':
	    fprintf(stdout, PROGRAM " " VERSION "\n");
	    exit(0);

	/* print help */
	case 'h':
	    fprintf(stderr,
PROGRAM " " VERSION ": .tap to native c64 tape mastering binary\n"
"Copyright (c) 2012, 2013 Daniel Kahlin <daniel@kahlin.net>\n"
"Written by Daniel Kahlin <daniel@kahlin.net>\n"
"\n"
"usage: " PROGRAM " [-o<name>][-t<str>][-f][-v][-d][-h][-V] <source>...\n"
"\n"
"Valid options:\n"
"    -T              title text (may be multiple)\n"
"    -o              output file\n"
"    -f              force reconversion even if cached copy exists\n"
"    -v              be verbose\n"
"    -d              debugging output\n"
"    -h              displays this help text\n"
"    -V              output program version\n"
	    );
	    exit(0);

	/* set output file */
	case 'o':
	    outname_g = optarg;
	    break;

	/* set title string */
	case 't':
	    title_g[titlenum_g] = optarg;
	    titlenum_g++;
	    break;

	/* set force mode */
	case 'f':
	    force_g = 1;
	    break;

	/* default behavior */
	default:
	    break;
	}
    }

    /*
     * optind now points at the first non option argument
     * we expect one more argument (inname)
     */
    int num_files = argc-optind;
    if (num_files < 1)
	panic("too few arguments");
    if (num_files > 1)
	panic("too many arguments");
    srcname = argv[optind];


    char tmpname[256];
    char exoname[256];
    snprintf(tmpname, 256, "%s.tmp", srcname);
    snprintf(exoname, 256, "%s.exo", srcname);

    if (force_g || (file_modified(srcname) > file_modified(exoname)) ) {
	tapfix(srcname, tmpname);
	exomize(tmpname, exoname);
	if (!debug_g) {
	    unlink(tmpname);
	}
    }


    /* write out the mastering program */
    fp = fopen(outname_g, "wb");
    if (!fp) {
	panic("couldn't open file '%s' for writing!", outname_g);
    }

    uint32_t sa = 0x0801;
    uint32_t ea;
    fwrite(data, 1, sizeof(data), fp);

    int i;
    for (i = 0; i < (4 - titlenum_g); i++) {
	fputc(13, fp);
    }
    fputc(129, fp); /* orange */
    fputc(32, fp);
    fputc(32, fp);
    for (i = 0; i < 36; i++) {
	fputc(184, fp);
    }

    fputc(158, fp); /* yellow */
    for (i = 0; i < titlenum_g; i++) {
	int j;
	int n = strlen(title_g[i]);
	fputc(13, fp);
	for (j = 0; j < ((40-n)/2); j++) {
	    fputc(' ', fp);
	}
	fprintf(fp, title_g[i]);
	fputc(13, fp);
    }

    fputc(129, fp); /* orange */
    fputc(32, fp);
    fputc(32, fp);
    for (i = 0; i < 36; i++) {
	fputc(185, fp);
    }
    fputc(0, fp);


    FILE *infp;
    infp = fopen(exoname, "rb");
    if (!infp) {
	panic("couldn't open '%s' for reading!", exoname);
    }
    while ( (c = fgetc(infp)) != EOF ) {
	fputc(c, fp);
    }
    fclose(infp);

    ea = sa + ftell(fp) - 2;

    fclose(fp);
    printf("output file range $%04X-$%04X\n", sa, ea);
    if (ea > 0x10000) {
	fprintf(stderr, "warning: file does not fit on a c64!\n");
    } else {
	if (ea > 0xd000) {
	    fprintf(stderr, "warning: not loadable without a custom loader! (>202 blocks)\n");
	}
    }

    exit(0);
}

time_t file_modified(const char *name)
{
    struct stat s;
    if ( stat(name, &s) == 0 ) {
	return s.st_mtime;
    }
    return 0;
}


#define FREQ 985248
void tapfix(const char *inname, const char *outname)
{
    FILE *infp, *outfp;
    int c;
    int t;

    infp = fopen(inname, "rb");
    if (!infp) {
	panic("couldn't open source file!");
    }
    outfp = fopen(outname, "wb");
    if (!outfp) {
	panic("couldn't open dest file!");
    }

    /* skip header */
    fseek(infp, 0x14, SEEK_CUR);

    t = 0;
    while ( (c = fgetc(infp)) != EOF ) {
	unsigned int val = c << 3;
	if ( c == 0 ) {
	    /* handle pause */
	    int i;

	    val = fgetc(infp);
	    val |= fgetc(infp) << 8;
	    val |= fgetc(infp) << 16;
	    printf ("t=%d: pause %d\n", t / FREQ, val);

	    for (i = 0; i < (val / 0x800); i++) {
		fputc(0xff, outfp);
	    }

	} else {
	    /* handle normal pulse */
	    unsigned char v, d;

	    d = c - 1;
	    v = (d & 0xc0) >> 6;
	    v |= (d & 0x3f) << 2;

	    fputc(v, outfp);
	}
	t += val;
    }
    /* end marker */
    fputc(0, outfp);

    printf ("processed %d seconds.\n", t / FREQ);

    fclose(infp);
    fclose(outfp);
}


void exomize(const char *inname, const char *outname)
{
    char cmd[256];
    time_t t1, t2;
    snprintf(cmd, 256, "exomizer raw -q -c -m2048 -C -M256 -o%s %s", outname, inname);
    //snprintf(cmd, 256, "exomizer raw -q -c -m2048 -o%s %s", outname, inname);
    printf("exomizing...");
    fflush(stdout);
    t1 = time(0);
    system(cmd);
    t2 = time(0);
    printf("%ds...ok\n", (int)(t2-t1));
}

/* eof */
